@@ -9,6 +9,7 @@ from . import sonos
 pp = pprint.PrettyPrinter(width=120)
 
 
+# Error handling functions
 def error_and_exit(msg):
     # Print to stderror
     print("Error:", msg, file=sys.stderr)
@@ -102,12 +103,11 @@ def list_queue(speaker, action, args, soco_function, use_local_speaker_list):
 def list_numbered_things(speaker, action, args, soco_function, use_local_speaker_list):
     if len(args) != 0:
         parameter_number_error(action, "no")
-        # Probably doesn't get here, but just in case
         return False
-    if soco_function == "get_sonos_favorites":
-        things = getattr(speaker.music_library, soco_function)()
+    if soco_function in ["get_sonos_favorites", "get_favorite_radio_stations", "get_albums", "get_artists", "get_tracks"]:
+        things = getattr(speaker.music_library, soco_function)(complete_result=True)
     else:
-        things = getattr(speaker, soco_function)()
+        things = getattr(speaker, soco_function)(complete_result=True)
     things_list = []
     for thing in things:
         things_list.append(thing.title)
@@ -115,7 +115,7 @@ def list_numbered_things(speaker, action, args, soco_function, use_local_speaker
     index = 0
     for thing in things_list:
         index += 1
-        print("{:3d}: {}".format(index, thing))
+        print("{:5d}: {}".format(index, thing))
     return True
 
 
@@ -155,7 +155,7 @@ def volume_actions(speaker, action, args, soco_function, use_local_speaker_list)
     return True
 
 
-def relative_volume_actions(speaker, action, args, soco_function, use_local_speaker_list):
+def relative_volume(speaker, action, args, soco_function, use_local_speaker_list):
     if len(args) != 1:
         parameter_number_error(action, "1")
         return False
@@ -165,6 +165,7 @@ def relative_volume_actions(speaker, action, args, soco_function, use_local_spea
         vol = int(args[0])
     except:
         parameter_type_error(action, "integer from -100 to 100")
+        return False
     if -100 <= vol <= 100:
         speaker.volume += vol
     else:
@@ -186,12 +187,15 @@ def print_info(speaker, action, args, soco_function, use_local_speaker_list):
 
 def playback_mode(speaker, action, args, soco_function, use_local_speaker_list):
     np = len(args)
+    if np not in [0, 1]:
+        parameter_number_error(action, "0 or 1")
+        return False
     possible_args = [
-            "normal",
-            "repeat_all",
-            "repeat_one",
-            "shuffle",
-            "shuffle_norepeat",
+        "normal",
+        "repeat_all",
+        "repeat_one",
+        "shuffle",
+        "shuffle_norepeat",
     ]
     if np == 0:
         print(speaker.play_mode)
@@ -200,9 +204,6 @@ def playback_mode(speaker, action, args, soco_function, use_local_speaker_list):
             speaker.play_mode = args[0]
         else:
             parameter_type_error(action, possible_args)
-    else:
-        parameter_number_error(action, "0 or 1")
-        return False
     return True
 
 
@@ -216,47 +217,92 @@ def transport_state(speaker, action, args, soco_function, use_local_speaker_list
 
 
 def play_favourite(speaker, action, args, soco_function, use_local_speaker_list):
-        if len(args) != 1:
-            parameter_number_error(action, "1")
-            return False
-        favourite = args[0].lower()
-        fs = speaker.music_library.get_sonos_favorites()
-        the_fav = None
-        # Strict match (case insensitive)
-        for f in fs:
-            if favourite == f.title.lower():
-                the_fav = f
-        # Loose substring match if strict match not available
-        if not the_fav:
-            for f in fs:
-                if favourite in f.title.lower():
-                    the_fav = f
-        if the_fav:
-            # play_uri works for some favourites
-            try:
-                uri = the_fav.get_uri()
-                metadata = the_fav.resource_meta_data
-                speaker.play_uri(uri=uri, meta=metadata)
-                return True  # Success
-            except Exception as e:
-                e1 = e
-                pass
-            # Other favourites have to be added to the queue, then played
-            try:
-                # speaker.clear_queue()
-                index = speaker.add_to_queue(the_fav)
-                speaker.play_from_queue(index, start=True)
-                return True
-            except Exception as e2:
-                error_and_exit("{}, {}".format(str(e1), str(e2)))
-                return False
-        error_and_exit("Favourite '{}' not found".format(favourite))
+    if len(args) != 1:
+        parameter_number_error(action, "1")
         return False
+    favourite = args[0]
+    fs = speaker.music_library.get_sonos_favorites()
+    the_fav = None
+    # Strict match
+    for f in fs:
+        if favourite == f.title:
+            the_fav = f
+            break
+    # Fuzzy match
+    favourite = favourite.lower()
+    if not the_fav:
+        for f in fs:
+            if favourite in f.title.lower():
+                the_fav = f
+                break
+    if the_fav:
+        # play_uri works for some favourites
+        try:
+            uri = the_fav.get_uri()
+            metadata = the_fav.resource_meta_data
+            speaker.play_uri(uri=uri, meta=metadata)
+            return True
+        except Exception as e:
+            e1 = e
+            pass
+        # Other favourites will be added to the queue, then played
+        try:
+            # Add to the end of the current queue and play
+            index = speaker.add_to_queue(the_fav, as_next=True)
+            speaker.play_from_queue(index, start=True)
+            return True
+        except Exception as e2:
+            error_and_exit("1: {} | 2:{}".format(str(e1), str(e2)))
+            return False
+    error_and_exit("Favourite '{}' not found".format(args[0]))
+    return False
+
+
+def play_favourite_radio(speaker, action, args, soco_function, use_local_speaker_list):
+    if len(args) != 1:
+        parameter_number_error(action, "1")
+        return False
+    favourite = args[0]
+    fs = speaker.music_library.get_favorite_radio_stations()
+    the_fav = None
+    # Strict match
+    for f in fs:
+        if favourite == f.title:
+            the_fav = f
+            break
+    # Fuzzy match
+    favourite = favourite.lower()
+    if not the_fav:
+        for f in fs:
+            if favourite in f.title.lower():
+                the_fav = f
+                break
+    if the_fav:
+        # play_uri works for some favourites
+        try:
+            uri = the_fav.get_uri()
+            metadata = the_fav.resource_meta_data
+            speaker.play_uri(uri=uri, meta=metadata)
+            return True
+        except Exception as e:
+            e1 = e
+            pass
+        # Other favourites will be added to the queue, then played
+        try:
+            # Add to the end of the current queue and play
+            index = speaker.add_to_queue(the_fav, as_next=True)
+            speaker.play_from_queue(index, start=True)
+            return True
+        except Exception as e2:
+            error_and_exit("1: {} | 2:{}".format(str(e1), str(e2)))
+            return False
+    error_and_exit("Favourite '{}' not found".format(args[0]))
+    return False
 
 
 def play_uri(speaker, action, args, soco_function, use_local_speaker_list):
     np = len(args)
-    if not (np == 1 or np == 2):
+    if np not in [1, 2]:
         parameter_number_error(action, "1 or 2")
         return False
     else:
@@ -272,6 +318,9 @@ def play_uri(speaker, action, args, soco_function, use_local_speaker_list):
 
 def sleep_timer(speaker, action, args, soco_function, use_local_speaker_list):
     np = len(args)
+    if np not in [0, 1]:
+        parameter_number_error(action, "0 or 1")
+        return False
     if np == 0:
         st = speaker.get_sleep_timer()
         if st:
@@ -287,9 +336,6 @@ def sleep_timer(speaker, action, args, soco_function, use_local_speaker_list):
             parameter_type_error(action, "integer > 0")
             return False
         speaker.set_sleep_timer(int(args[0]))
-    else:
-        parameter_number_error(action, "0 or 1")
-        return False
     return True
 
 
@@ -334,6 +380,9 @@ def zones(speaker, action, args, soco_function, use_local_speaker_list):
 
 def play_from_queue(speaker, action, args, soco_function, use_local_speaker_list):
     np = len(args)
+    if np not in [0, 1]:
+        parameter_number_error(action, "0 or 1")
+        return False
     if np == 0:
         speaker.play_from_queue(0)
     elif np == 1:
@@ -347,9 +396,6 @@ def play_from_queue(speaker, action, args, soco_function, use_local_speaker_list
         else:
             error_and_exit("Queue index '{}' is out of range".format(index))
             return False
-    else:
-        parameter_number_error(action, "0 or 1")
-        return False
     return True
 
 
@@ -364,7 +410,7 @@ def remove_from_queue(speaker, action, args, soco_function, use_local_speaker_li
         return False
     qs = speaker.queue_size
     if 1 <= index <= qs:
-        speaker.play_from_queue(index - 1)
+        speaker.remove_from_queue(index - 1)
     else:
         error_and_exit("Queue index should be between 1 and {}".format(qs))
         return False
@@ -391,21 +437,54 @@ def seek(speaker, action, args, soco_function, use_local_speaker_list):
     return True
 
 
-def add_playlist_to_queue(speaker, action, args, soco_function, use_local_speaker_list):
+def playlist_operations(speaker, action, args, soco_function, use_local_speaker_list):
     if len(args) != 1:
         parameter_number_error(action, "1")
         return False
-    name = args[0].lower()
+    name = args[0]
+    if soco_function == "create_sonos_playlist":
+        getattr(speaker, soco_function)(name)
+        return True
+    if soco_function == "add_uri_to_queue":
+        getattr(speaker, soco_function)(name)
+        return True
     playlists = speaker.get_sonos_playlists()
     # Strict match
     for playlist in playlists:
-        if name == playlist.title.lower():
-            speaker.add_to_queue(playlist)
+        if name == playlist.title:
+            getattr(speaker, soco_function)(playlist)
             return True
     # Fuzzy match
+    name = name.lower()
     for playlist in playlists:
         if name in playlist.title.lower():
-            speaker.add_to_queue(playlist)
+            getattr(speaker, soco_function)(playlist)
+            return True
+    error_and_exit("Playlist {} not found".format(args[0]))
+    return False
+
+
+def remove_from_playlist(speaker, action, args, soco_function, use_local_speaker_list):
+    if len(args) != 2:
+        parameter_number_error(action, "2")
+        return False
+    name = args[0]
+    try:
+        track_number = int(args[1])
+    except:
+        parameter_type_error(action, "integer (track number)")
+        return False
+    playlists = speaker.get_sonos_playlists()
+    # Strict match
+    for playlist in playlists:
+        if name == playlist.title:
+            getattr(speaker, soco_function)(playlist, track_number - 1)
+            return True
+    # Fuzzy match
+    name = name.lower()
+    for playlist in playlists:
+        if name in playlist.title.lower():
+            getattr(speaker, soco_function)(playlist, track_number - 1)
             return True
     error_and_exit("Playlist {} not found".format(args[0]))
     return False
@@ -433,7 +512,7 @@ def line_in(speaker, action, args, soco_function, use_local_speaker_list):
 
 def eq(speaker, action, args, soco_function, use_local_speaker_list):
     np = len(args)
-    if np > 1:
+    if np not in [0, 1]:
         parameter_number_error(action, "0 or 1")
         return False
     if np == 0:
@@ -454,12 +533,12 @@ def eq(speaker, action, args, soco_function, use_local_speaker_list):
 
 def balance(speaker, action, args, soco_function, use_local_speaker_list):
     np = len(args)
-    if np > 1:
+    if np not in [0, 1]:
         parameter_number_error(action, "0 or 1")
         return False
     if np == 0:
         left, right = getattr(speaker, soco_function)
-        # Convert to something more intelligible than a tuple
+        # Convert to something more intelligible than a 2-tuple
         # Use range from -100 (full left) to +100 (full right)
         print(right - left)
     elif np == 1:
@@ -469,8 +548,12 @@ def balance(speaker, action, args, soco_function, use_local_speaker_list):
             parameter_type_error(action, "integer from -100 to 100")
             return False
         if -100 <= setting <= 100:
-            left = 0 - setting
-            right = 0 + setting
+            if setting >= 0:
+                left = 100 - setting
+                right = 100
+            elif setting < 0:
+                left = 100
+                right = 100 + setting
             setattr(speaker, soco_function, (left, right))
         else:
             parameter_type_error(action, "integer from -100 to 100")
@@ -487,6 +570,9 @@ def reindex(speaker, action, args, soco_function, use_local_speaker_list):
 
 
 def info(speaker, action, args, soco_function, use_local_speaker_list):
+    if len(args) != 0:
+        parameter_number_error(action, "no")
+        return False
     info = speaker.get_speaker_info()
     model = info["model_name"].lower()
     if not ("boost" in model or "bridge" in model):
@@ -510,6 +596,8 @@ def info(speaker, action, args, soco_function, use_local_speaker_list):
         info["night_mode"] = speaker.night_mode
         info["is_soundbar"] = speaker.is_soundbar
         info["is_playing_line_in"] = speaker.is_playing_line_in
+        info["is_playing_radio"] = speaker.is_playing_radio
+        info["is_playing_tv"] = speaker.is_playing_tv
         info["is_visible"] = speaker.is_visible
     for item in sorted(info):
         print("  {} = {}".format(item, info[item]))
@@ -525,8 +613,7 @@ def groups(speaker, action, args, soco_function, use_local_speaker_list):
             print("[{}] : ".format(group.short_label), end="")
             for member in group.members:
                 print(
-                    "{} ({}) ".format(member.player_name, member.ip_address),
-                    end="",
+                    "{} ({}) ".format(member.player_name, member.ip_address), end="",
                 )
             print()
     return True
@@ -536,11 +623,7 @@ def process_action(speaker, action, args, use_local_speaker_list):
     sonos_function = actions.get(action, None)
     if sonos_function:
         return sonos_function.processing_function(
-            speaker,
-            action,
-            args,
-            sonos_function.soco_function,
-            use_local_speaker_list,
+            speaker, action, args, sonos_function.soco_function, use_local_speaker_list,
         )
     else:
         return False
@@ -548,9 +631,7 @@ def process_action(speaker, action, args, use_local_speaker_list):
 
 # Type for holding action processing functions
 SonosFunction = namedtuple(
-    "SonosFunction",
-    ["processing_function", "soco_function",],
-    rename=False,
+    "SonosFunction", ["processing_function", "soco_function",], rename=False,
 )
 
 # Actions and associated processing functions
@@ -591,18 +672,20 @@ actions = {
     "gv": SonosFunction(volume_actions, "group_volume"),
     "ramp_to_volume": SonosFunction(volume_actions, "ramp_to_volume"),
     "ramp": SonosFunction(volume_actions, "ramp_to_volume"),
-    "relative_volume": SonosFunction(relative_volume_actions, "relative_volume"),
-    "rel_vol": SonosFunction(relative_volume_actions, "relative_volume"),
-    "rv": SonosFunction(relative_volume_actions, "relative_volume"),
-    "group_relative_volume": SonosFunction(relative_volume_actions, "group_relative_volume"),
-    "group_rel_vol": SonosFunction(relative_volume_actions, "group_relative_volume"),
-    "grv": SonosFunction(relative_volume_actions, "group_relative_volume"),
+    "relative_volume": SonosFunction(relative_volume, "relative_volume"),
+    "rel_vol": SonosFunction(relative_volume, "relative_volume"),
+    "rv": SonosFunction(relative_volume, "relative_volume"),
+    "group_relative_volume": SonosFunction(relative_volume, "group_relative_volume"),
+    "group_rel_vol": SonosFunction(relative_volume, "group_relative_volume"),
+    "grv": SonosFunction(relative_volume, "group_relative_volume"),
     "track": SonosFunction(print_info, "get_current_track_info"),
     "play_mode": SonosFunction(playback_mode, "play_mode"),
     "mode": SonosFunction(playback_mode, "play_mode"),
     "playback_state": SonosFunction(transport_state, "get_current_transport_info"),
     "playback": SonosFunction(transport_state, "get_current_transport_info"),
     "state": SonosFunction(transport_state, "get_current_transport_info"),
+    "play_favourite": SonosFunction(play_favourite, "play_favorite"),
+    "play_favorite": SonosFunction(play_favourite, "play_favorite"),
     "favourite": SonosFunction(play_favourite, "play_favorite"),
     "favorite": SonosFunction(play_favourite, "play_favorite"),
     "play_fav": SonosFunction(play_favourite, "play_favorite"),
@@ -638,9 +721,10 @@ actions = {
     "save_queue": SonosFunction(save_queue, "create_sonos_playlist_from_queue"),
     "sq": SonosFunction(save_queue, "create_sonos_playlist_from_queue"),
     "queue_length": SonosFunction(no_args_one_output, "queue_size"),
-    "add_playlist_to_queue": SonosFunction(add_playlist_to_queue, "add_to_queue"),
-    "add_pl_to_queue": SonosFunction(add_playlist_to_queue, "add_to_queue"),
-    "apq": SonosFunction(add_playlist_to_queue, "add_to_queue"),
+    "ql": SonosFunction(no_args_one_output, "queue_size"),
+    "add_playlist_to_queue": SonosFunction(playlist_operations, "add_to_queue"),
+    "add_pl_to_queue": SonosFunction(playlist_operations, "add_to_queue"),
+    "apq": SonosFunction(playlist_operations, "add_to_queue"),
     "pause_all": SonosFunction(operate_on_all, "pause"),
     "seek": SonosFunction(seek, "seek"),
     "line_in": SonosFunction(line_in, ""),
@@ -652,4 +736,20 @@ actions = {
     "groups": SonosFunction(groups, "groups"),
     "pair": SonosFunction(group_or_pair, "create_stereo_pair"),
     "unpair": SonosFunction(no_args_no_output, "separate_stereo_pair"),
+    "delete_playlist": SonosFunction(playlist_operations, "remove_sonos_playlist"),
+    "remove_playlist": SonosFunction(playlist_operations, "remove_sonos_playlist"),
+    "clear_playlist": SonosFunction(playlist_operations, "clear_sonos_playlist"),
+    "create_playlist": SonosFunction(playlist_operations, "create_sonos_playlist"),
+    "add_uri_to_queue": SonosFunction(playlist_operations, "add_uri_to_queue"),
+    "auq": SonosFunction(playlist_operations, "add_uri_to_queue"),
+    "remove_from_playlist": SonosFunction(remove_from_playlist, "remove_from_sonos_playlist"),
+    "rfp": SonosFunction(remove_from_playlist, "remove_from_sonos_playlist"),
+    "favorite_radio_stations": SonosFunction(list_numbered_things, "get_favorite_radio_stations"),
+    "favourite_radio_stations": SonosFunction(list_numbered_things, "get_favorite_radio_stations"),
+    "play_favourite_radio_station": SonosFunction(play_favourite_radio, "play_uri"),
+    "play_favorite_radio_station": SonosFunction(play_favourite_radio, "play_uri"),
+    "pfrs": SonosFunction(play_favourite_radio, "play_uri"),
+    "albums": SonosFunction(list_numbered_things, "get_albums"),
+    "artists": SonosFunction(list_numbered_things, "get_artists"),
+    "tracks": SonosFunction(list_numbered_things, "get_tracks"),
 }
